from autohack.core.util import *
from typing import Any
import logging, pathlib, json, os


class Config:
    def __init__(
        self,
        configFilePath: pathlib.Path,
        defaultConfig: dict[str, Any],
        logger: logging.Logger,
    ) -> None:
        self.defaultConfig = defaultConfig
        self.configFilePath = configFilePath
        self.logger = logger
        self.logger.info(f'[config] Config file path: "{self.configFilePath}"')
        self.config = self.loadConfig()

    def loadConfig(self) -> dict[str, Any]:
        if not os.path.exists(self.configFilePath):
            json.dump(
                self.defaultConfig,
                open(self.configFilePath, "w", encoding="utf-8"),
                indent=4,
            )
            self.logger.info("[config] Config file created.")
            write(f"Config file created at {self.configFilePath}.")
            exitProgram(0)

        with open(self.configFilePath, "r", encoding="utf-8") as configFile:
            config = json.load(configFile)

        if self.defaultConfig["_version"] > config.get("_version", 0):
            mergedConfig = self.mergeConfigs(config, self.defaultConfig)
            mergedConfig["_version"] = self.defaultConfig["_version"]
            json.dump(
                mergedConfig, open(self.configFilePath, "w", encoding="utf-8"), indent=4
            )
            write(
                f"Config file {self.configFilePath} updated to version {self.defaultConfig['_version']}.",
                2,
            )
            self.logger.info("[config] Config file updated.")
            config = mergedConfig

        self.logger.info("[config] Config file loaded.")
        return config

    def mergeConfigs(
        self, old: dict[str, Any], newDefault: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Merge the old config with the new default config.
        - If a key exists in both, the value from the old config is used.
        - If a key exists only in the new default config, it is added.
        - If a key exists only in the old config, it is ignored.
        """
        merged = {}
        for key in newDefault:
            if (
                key in old
                and isinstance(newDefault[key], dict)
                and isinstance(old[key], dict)
            ):
                merged[key] = self.mergeConfigs(old[key], newDefault[key])
            else:
                merged[key] = old.get(key, newDefault[key])
        return merged

    def getConfigEntry(self, entryName: str) -> Any:
        entryTree = entryName.split(".")
        result = self.config

        for entryItem in entryTree:
            result = result.get(entryItem, None)
            if result is None:
                break

        self.logger.debug(f'[config] Get config entry: "{entryName}" = "{result}"')
        return result

    def modifyConfigEntry(self, entryName: str, newValue: Any) -> bool:
        """Returns True if the entry was modified, False if it does not exist."""
        entryTree = entryName.split(".")
        currentLevel = self.config

        for level in entryTree[:-1]:
            if not isinstance(currentLevel, dict) or level not in currentLevel:
                return False
            currentLevel = currentLevel[level]
        lastLevel = entryTree[-1]
        if not isinstance(currentLevel, dict) or lastLevel not in currentLevel:
            return False
        currentLevel[lastLevel] = newValue

        json.dump(
            self.config, open(self.configFilePath, "w", encoding="utf-8"), indent=4
        )
        self.logger.debug(f'[config] Modify entry: "{entryName}" = "{newValue}"')
        return True



import pathlib, time, sys, os


def ensureDirExists(dirPath: pathlib.Path) -> None:
    if not os.path.exists(dirPath):
        os.makedirs(dirPath)


def mswindows() -> bool:
    try:
        import msvcrt
    except ModuleNotFoundError:
        return False
    else:
        return True


def formatTime(t: time.struct_time = time.localtime()) -> str:
    return time.strftime("%Y%m%d%H%M%S", t)


def writeData(filePath: pathlib.Path, data: bytes) -> None:
    ensureDirExists(filePath.parent)
    open(filePath, "wb").write(data)


def readData(filePath: pathlib.Path) -> bytes:
    return open(filePath, "rb").read()


def clearLine() -> None:
    write("\x1b[2K\r")


def prevLine() -> None:
    write("\x1b[1A")


def outputEndl(count: int = 1) -> None:
    sys.stdout.write("\n" * count)


def write(message: str, endl: int = 0, clear: bool = False) -> None:
    if clear:
        clearLine()
    sys.stdout.write(message)
    outputEndl(endl)
    sys.stdout.flush()


def hideCursor() -> None:
    # https://www.cnblogs.com/chargedcreeper/p/-/ANSI
    write("\x1b[?25l")


def showCursor() -> None:
    write("\x1b[?25h")


def highlightText(message: str) -> str:
    return f"\x1b[1;31m{message}\x1b[0m"


def exitProgram(exitCode: int = 0, pure: bool = False) -> None:
    if not pure:
        showCursor()
    sys.exit(exitCode)



from autohack.core.constant import *
from autohack.core.exception import *
from autohack.core.path import *
from autohack.core.run import *
from autohack.core.util import *
from typing import Any, Callable, TypeAlias
import importlib.util, subprocess, inspect, pathlib

checkerType: TypeAlias = Callable[[bytes, bytes, bytes, dict], tuple[bool, str]]
activateType: TypeAlias = Callable[[dict], checkerType]


def builtinBasicCheckerActivate(args: dict) -> checkerType:
    def builtinBasicChecker(
        input: bytes, output: bytes, answer: bytes, args: dict
    ) -> tuple[bool, str]:
        outputStr = output.decode().rstrip("\n")
        answerStr = answer.decode().rstrip("\n")
        outputLines = outputStr.splitlines()
        answerLines = answerStr.splitlines()
        if len(outputLines) != len(answerLines):
            return (False, "Output and answer have different number of lines.")
        for i in range(len(outputLines)):
            if outputLines[i].rstrip() != answerLines[i].rstrip():
                return (False, f"Line {i + 1} does not match.")
        return (True, "Output matches the answer.")

    return builtinBasicChecker


def builtinAlwaysACCheckerActivate(args: dict) -> checkerType:
    def builtinAlwaysACChecker(
        input: bytes, output: bytes, answer: bytes, args: dict
    ) -> tuple[bool, str]:
        return (True, "Always AC checker.")

    return builtinAlwaysACChecker


"""
代码	        返回值	含义
_ok	            0	    正确
_wa	            1	    错误
_pe	            2	    格式错误
_fail	        3	    运行失败，程序出错
_dirt	        4	    输出文件含有多余信息
_points	        5	    部分分数
_unexpected_eof	8	    文件读完时仍然尝试读入
_partially	    16+分数	部分正确

from https://www.luogu.com/article/t5rrziq7
"""


def builtinTestlibCheckerActivate(args: dict) -> checkerType:
    ensureDirExists(DATA_FOLDER_PATH / "testlibCheckerCache")
    checkerPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "checker"
    compileCommand = [
        args.get("compiler", "g++"),
        args.get("checker", "checker.cpp"),
        "-o",
        checkerPath.as_posix(),
    ]
    compileCommand += args.get("compile_args", [])
    try:
        compileCode(compileCommand, "checker")
    except CompilationError as e:
        print(e.getMessage())
        raise

    def builtinTestlibChecker(
        input: bytes, output: bytes, answer: bytes, args: dict
    ) -> tuple[bool, str]:
        inputPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "input"
        outputPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "output"
        answerPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "answer"
        resultPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "result"
        checkerPath = DATA_FOLDER_PATH / "testlibCheckerCache" / "checker"
        writeData(inputPath, input)
        writeData(outputPath, output)
        writeData(answerPath, answer)
        command = [
            checkerPath.as_posix(),
            inputPath.as_posix(),
            outputPath.as_posix(),
            answerPath.as_posix(),
            resultPath.as_posix(),
        ]
        result = subprocess.run(
            command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        ).returncode
        if not resultPath.exists():
            raise FileNotFoundError("Testlib checker did not produce a result file.")
        resultContent = readData(resultPath).decode().strip()
        if result == 0:
            return (True, f"{resultContent} (Code {result})")
        elif result == 3:
            raise RuntimeError(
                f"Testlib checker runtime error. Checker output: {resultContent}"
            )
        return (False, f"{resultContent} (Code {result})")

    return builtinTestlibChecker


BUILTIN = [
    ("builtin_basic", builtinBasicCheckerActivate),
    ("builtin_always_ac", builtinAlwaysACCheckerActivate),
    ("builtin_testlib", builtinTestlibCheckerActivate),
]


"""
Checker 中的 activate 函数签名为 (dict) -> Callable[[bytes, bytes, dict], tuple[bool, str]]
即接受 args 返回 checker 函数
"""


def getChecker(
    checkerFolder: pathlib.Path, checkerName: str, args: dict[str, Any]
) -> checkerType:
    # 如果 checkerName 在 BUILTIN 中，直接返回对应的函数
    for name, func in BUILTIN:
        if name == checkerName:
            return func(args)

    checkerPath = checkerFolder / f"{checkerName}.py"
    if not checkerPath.exists():
        raise FileNotFoundError(f'Checker "{checkerPath}" not found.')

    spec = importlib.util.spec_from_file_location(checkerName, checkerPath)
    if spec is None or spec.loader is None:
        raise ImportError(f'Could not load spec for checker "{checkerName}".')
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if not hasattr(module, "activate"):
        raise AttributeError(
            f"Checker '{checkerName}' does not have a 'activate' function."
        )

    try:
        checker = module.activate(args)
    except Exception as e:
        raise RuntimeError(f"Error while activating checker '{checkerName}': {e}")

    sig = inspect.signature(checker)
    params = sig.parameters
    param_types = [param.annotation for param in params.values()]

    if param_types != [bytes, bytes, bytes, dict]:
        raise TypeError(
            f"Checker '{checkerName}' function parameters must be of types (bytes, bytes, dict)."
        )

    if sig.return_annotation != tuple[bool, str]:
        raise TypeError(
            f"Checker '{checkerName}' function must return tuple[bool, str]."
        )

    return checker



from autohack.core.checker import *
from autohack.core.constant import *
from autohack.core.exception import *
from autohack.core.path import *
from autohack.core.util import *
from autohack.core.run import *
from autohack.lib.config import *
from autohack.lib.logger import *
from typing import Callable
import traceback, argparse, colorama, logging, time, uuid, os

CLIENT_ID = str(uuid.uuid4())
LOG_TIME = time.localtime()


def main() -> None:
    global CLIENT_ID, LOG_TIME

    argsParser = argparse.ArgumentParser(
        prog="autohack", description="autohack-next - Automated hack data generator"
    )
    argsParser.add_argument(
        "--version", action="store_true", help="Show version information"
    )
    argsParser.add_argument("--version-id", action="store_true", help="Show version ID")
    argsParser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode with DEBUG logging level",
    )
    # TODO: 添加一个参数用于清除过往数据

    args = argsParser.parse_args()

    if args.version:
        write(f"{VERSION}")
        exitProgram(0, True)

    if args.version_id:
        write(f"{VERSION_ID}")
        exitProgram(0, True)

    hideCursor()

    if args.debug:
        write("Debug mode enabled. Logging level set to DEBUG.", 2)

    ensureDirExists(CHECKER_FOLDER_PATH)
    ensureDirExists(LOG_FOLDER_PATH)

    loggerObj = Logger(
        LOG_FOLDER_PATH, logging.DEBUG if args.debug else logging.INFO, LOG_TIME
    )
    logger = loggerObj.getLogger()

    config = Config(CONFIG_FILE_PATH, DEFAULT_CONFIG, logger)

    logger.info(f'[autohack] Data folder path: "{DATA_FOLDER_PATH}"')
    logger.info(f"[autohack] Client ID: {CLIENT_ID}")
    logger.info(f"[autohack] Initialized. Version: {VERSION}")
    write(f"autohack-next {VERSION} - Client ID: {CLIENT_ID}", 2)
    write(
        f"Hack data storaged to {getHackDataStorageFolderPath(CLIENT_ID, LOG_TIME)}", 1
    )
    write(f"Log file: {loggerObj.getLogFilePath()}", 1)
    write(f"Error export to {getExportFolderPath(LOG_TIME, CLIENT_ID)}", 1)
    write(f"Custom checker folder: {CHECKER_FOLDER_PATH}", 2)

    write("Activating checker...")

    currentChecker: Callable[[bytes, bytes, bytes, dict], tuple[bool, str]] = (
        lambda l, o, a, ar: (False, "No checker activated.")
    )

    try:
        currentChecker = getChecker(
            CHECKER_FOLDER_PATH,
            config.getConfigEntry("checker.name"),
            config.getConfigEntry("checker.args"),
        )
    except Exception as e:
        logger.critical(f"[autohack] {e}")
        write("Checker activation failed.", 1, True)
        write(highlightText(e.__str__()))
        exitProgram(1)

    write("Checker activated.", 2, True)

    for i in range(WAIT_TIME_BEFORE_START):
        write(f"Starting in {WAIT_TIME_BEFORE_START-i} seconds...", clear=True)
        time.sleep(1)

    fileList = [
        [config.getConfigEntry("commands.compile.source"), "source code"],
        [config.getConfigEntry("commands.compile.std"), "standard code"],
        [config.getConfigEntry("commands.compile.generator"), "generator code"],
    ]
    for file in fileList:
        write(f"Compile {file[1]}.", clear=True)
        try:
            compileCode(file[0], file[1])
        except CompilationError as e:
            logger.error(
                f"[autohack] {file[1].capitalize()} compilation failed with return code {e.returnCode} and message:\n{e.message.decode()}"
            )
            write(highlightText(e.__str__()), 2, True)
            write(e.getMessage())
            exitProgram(1)
        else:
            logger.debug(f"[autohack] {file[1].capitalize()} compiled successfully.")

    write("Compile finished.", 2, True)

    dataCount, errorDataCount = 0, 0
    lastStatusError = False
    generateCommand = config.getConfigEntry("commands.run.generator")
    stdCommand = config.getConfigEntry("commands.run.std")
    sourceCommand = config.getConfigEntry("commands.run.source")
    timeLimit = config.getConfigEntry("time_limit") / 1000
    memoryLimit = config.getConfigEntry("memory_limit") * 1024 * 1024
    inputFilePath = config.getConfigEntry("paths.input")
    answerFilePath = config.getConfigEntry("paths.answer")
    outputFilePath = config.getConfigEntry("paths.output")
    maximumDataLimit = config.getConfigEntry("maximum_number_of_data")
    errorDataLimit = config.getConfigEntry("error_data_number_limit")
    refreshSpeed = config.getConfigEntry("refresh_speed")
    checkerArgs = config.getConfigEntry("checker.args")

    startTime = time.time()

    while (maximumDataLimit <= 0 or dataCount < maximumDataLimit) and (
        errorDataLimit <= 0 or errorDataCount < errorDataLimit
    ):
        dataInput = b""
        dataAnswer = b""

        dataCount += 1

        try:
            write(f"{dataCount}: Generate input.", clear=True)
            logger.debug(f"[autohack] Generating data {dataCount}.")
            dataInput = generateInput(generateCommand, CLIENT_ID)
        except InputGenerationError as e:
            logger.error(
                f"[autohack] Input generation failed with return code {e.returnCode}."
            )
            write(highlightText(e.__str__()), 1, True)
            inputExportPath = getExportDataPath(
                getExportFolderPath(LOG_TIME, CLIENT_ID), "input"
            )
            writeData(inputExportPath, dataInput)
            write(highlightText(f"Input data saved to {inputExportPath}"), clear=True)
            exitProgram(1)

        try:
            write(f"{dataCount}: Generate answer.", clear=True)
            logger.debug(f"[autohack] Generating answer for data {dataCount}.")
            dataAnswer = generateAnswer(
                stdCommand,
                dataInput,
                CLIENT_ID,
            )
        except AnswerGenerationError as e:
            logger.error(
                f"[autohack] Answer generation failed with return code {e.returnCode}."
            )
            write(highlightText(e.__str__()), 1, True)
            inputExportPath = getExportDataPath(
                getExportFolderPath(LOG_TIME, CLIENT_ID), "input"
            )
            writeData(inputExportPath, dataInput)
            write(highlightText(f"Input data saved to {inputExportPath}"), 1, True)
            answerExportPath = getExportDataPath(
                getExportFolderPath(LOG_TIME, CLIENT_ID), "answer"
            )
            writeData(answerExportPath, dataAnswer)
            write(highlightText(f"Answer data saved to {answerExportPath}"), clear=True)
            exitProgram(1)

        write(f"{dataCount}: Run source code.", clear=True)
        logger.debug(f"[autohack] Run source code for data {dataCount}.")
        result = runSourceCode(sourceCommand, dataInput, timeLimit, memoryLimit)
        if result.stdout is None:
            result.stdout = b""
        if result.stderr is None:
            result.stderr = b""

        # TODO: Refresh when running exe. Use threading or async?
        if dataCount % refreshSpeed == 0 or lastStatusError:
            lastStatusError = False
            currentTime = time.time()
            outputEndl()
            write(
                f"Time taken: {currentTime - startTime:.2f} seconds, average {dataCount/(currentTime - startTime):.2f} data per second, {(currentTime - startTime)/dataCount:.2f} second per data.{f" ({dataCount*100/maximumDataLimit:.0f}%)" if maximumDataLimit > 0 else ""}",
                clear=True,
            )
            prevLine()

        saveData, termMessage, logMessage, extMessage, exitAfterSave = (
            False,
            "",
            "",
            None,
            False,
        )

        if result.memoryOut:
            saveData = True
            termMessage = logMessage = f"Memory limit exceeded for data {dataCount}."
        elif result.timeOut:
            saveData = True
            termMessage = logMessage = f"Time limit exceeded for data {dataCount}."
        elif result.returnCode != 0:
            saveData = True
            termMessage = logMessage = (
                f"Runtime error for data {dataCount} with return code {result.returnCode}."
            )

        checkerResult = (False, "Checker not executed.")
        try:
            checkerResult = currentChecker(
                dataInput, result.stdout, dataAnswer, checkerArgs
            )
        except Exception as e:
            saveData = True
            termMessage = f"Checker error for data {dataCount}."
            logMessage = f"Checker error for data {dataCount}. Exception: {e}"
            extMessage = f"Traceback:\n{traceback.format_exc()}"
            checkerResult = (False, "Checker exception occurred.")
            exitAfterSave = True

        if not saveData and not checkerResult[0]:
            saveData = True
            termMessage = f"Wrong answer for data {dataCount}."
            logMessage = (
                f"Wrong answer for data {dataCount}. Checker output: {checkerResult[1]}"
            )
            extMessage = checkerResult[1]

        if saveData:
            lastStatusError = True
            errorDataCount += 1
            writeData(
                getHackDataFilePath(
                    getHackDataStorageFolderPath(CLIENT_ID, LOG_TIME),
                    errorDataCount,
                    inputFilePath,
                ),
                dataInput,
            )
            writeData(
                getHackDataFilePath(
                    getHackDataStorageFolderPath(CLIENT_ID, LOG_TIME),
                    errorDataCount,
                    answerFilePath,
                ),
                dataAnswer,
            )
            writeData(
                getHackDataFilePath(
                    getHackDataStorageFolderPath(CLIENT_ID, LOG_TIME),
                    errorDataCount,
                    outputFilePath,
                ),
                result.stdout,
            )
            write(f"[{errorDataCount}]: {termMessage}", 1, True)
            if extMessage is not None:
                write(f"{(len(f'[{errorDataCount}]: ')-3)*' '} - {extMessage}", 1, True)
            logger.info(f"[autohack] {logMessage}")

        if exitAfterSave:
            write("Exiting due to checker exception.", clear=True)
            exitProgram(0)

    endTime = time.time()

    write(
        f"Finished. {dataCount} data generated, {errorDataCount} error data found.",
        1,
        True,
    )
    write(
        f"Time taken: {endTime - startTime:.2f} seconds, average {dataCount/(endTime - startTime):.2f} data per second, {(endTime - startTime)/dataCount:.2f} second per data.",
        1,
        True,
    )

    # if errorDataCount == 0:
    #     shutil.rmtree(getHackDataStorageFolderPath(clientID))
    #     write("No error data found. Hack data folder removed.", 1)
    #     logger.info("[autohack] No error data found. Hack data folder removed.")

    if (
        HACK_DATA_STORAGE_FOLDER_PATH.exists()
        and HACK_DATA_STORAGE_FOLDER_PATH.stat().st_size > DATA_FOLDER_MAX_SIZE
    ):
        logger.warning(
            f"[autohack] Hack data storage folder size exceeds 256 MB: {HACK_DATA_STORAGE_FOLDER_PATH}"
        )
        write(
            f"Warning: Hack data storage folder size exceeds 256 MB: {HACK_DATA_STORAGE_FOLDER_PATH}",
            1,
        )

    write("Executing post process command.", 1)
    os.system(config.getConfigEntry("command_at_end"))
    logger.info("[autohack] Finished.")


if __name__ == "__main__" or os.getenv("AUTOHACK_ENTRYPOINT", "0") == "1":
    colorama.just_fix_windows_console()

    try:
        main()

    except KeyboardInterrupt:
        write(highlightText("Process interrupted by user."))

    except Exception as e:
        write(highlightText(f"Unhandled exception."), 1)

        errorFilePath = getExportFolderPath(LOG_TIME, CLIENT_ID) / f"error.log"
        ensureDirExists(errorFilePath.parent)
        errorFile = open(errorFilePath, "w+", encoding="utf-8")
        traceback.print_exc(file=errorFile)
        errorFile.close()

        write(highlightText(f"Error details saved to {errorFilePath}"), 2)
        # logger.critical(f"[autohack] Unhandled exception.")

        traceback.print_exc()
        exitProgram(1)

    exitProgram(0)



import logging, pathlib, json, os


class I18N:
    def __init__(
        self,
        translationFileDir: pathlib.Path,
        language: str,
        logger: logging.Logger,
    ) -> None:
        self.translationFile = os.path.join(translationFileDir, f"{language}.json")
        self.logger = logger
        self.logger.info(f'[i18n] Translation file: "{self.translationFile}"')
        self.translations = self.loadTranslationFile()

    def loadTranslationFile(self) -> dict[str, str]:
        if not os.path.exists(self.translationFile):
            self.logger.critical("[i18n] Translation file not found.")
            raise FileNotFoundError(
                f"Translation file {self.translationFile} not found."
            )
        with open(self.translationFile, "r", encoding="utf-8") as translationFile:
            translations = json.load(translationFile)

        self.logger.info("[i18n] Translation file loaded.")
        return translations

    def translate(self, key: str) -> str:
        return self.translations.get(key, key)



from autohack.core.exception import *
import subprocess, threading, psutil, time


class CodeRunner:
    class Result:
        def __init__(
            self,
            timeOut: bool,
            memoryOut: bool,
            returnCode: int | None,
            stdout: bytes | None,
            stderr: bytes | None,
        ) -> None:
            self.timeOut = timeOut
            self.memoryOut = memoryOut
            self.returnCode = returnCode
            self.stdout = stdout
            self.stderr = stderr

    def __init__(self):
        self.timeOut = False
        self.memoryOut = False

    def memoryMonitor(
        self, pid: int, timeLimit: float | None, memoryLimit: int | None
    ) -> None:
        try:
            psutilProcess = psutil.Process(pid)
            startTime = psutilProcess.create_time()
            while True:
                # psutilProcess.cpu_times();
                if timeLimit is not None and time.time() - startTime > timeLimit:
                    self.timeOut = True
                    psutilProcess.kill()
                    return
                # 测出来是资源监视器内存中提交那栏 *1024
                if (
                    memoryLimit is not None
                    and psutilProcess.memory_info().vms > memoryLimit
                ):
                    self.memoryOut = True
                    psutilProcess.kill()
                    return
        except psutil.NoSuchProcess:
            return

    def run(
        self,
        *popenargs,
        inputContent: bytes | None = None,
        timeLimit: float | None = None,
        memoryLimit: int | None = None,
        **kwargs,
    ) -> Result:
        returnCode = 0
        stdout = None
        stderr = None
        with subprocess.Popen(*popenargs, **kwargs) as process:
            monitor = threading.Thread(
                target=self.memoryMonitor,
                args=(
                    process.pid,
                    timeLimit,
                    memoryLimit,
                ),
            )
            monitor.start()
            stdout, stderr = process.communicate(inputContent)  # type: ignore
            returnCode = process.poll()
        return self.Result(self.timeOut, self.memoryOut, returnCode, stdout, stderr)  # type: ignore


def compileCode(compileCommand: list, fileName: str) -> None:
    try:
        process = subprocess.Popen(
            compileCommand, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    except OSError:
        return
    output = process.communicate()[0]
    if process.returncode != 0:
        raise CompilationError(fileName, output, process.returncode)


def generateInput(generateCommand: list, clientID: str) -> bytes:
    try:
        process = subprocess.Popen(
            generateCommand, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )
    except OSError:
        return b""
    dataInput = process.communicate()[0]
    if process.returncode != 0:
        raise InputGenerationError(clientID, process.returncode)
    return dataInput


def generateAnswer(generateCommand: list, dataInput: bytes, clientID: str) -> bytes:
    try:
        process = subprocess.Popen(
            generateCommand,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
    except OSError:
        return b""
    dataAnswer = process.communicate(dataInput)[0]
    if process.returncode != 0:
        raise AnswerGenerationError(clientID, process.returncode)
    return dataAnswer


def runSourceCode(
    runCommand: list, dataInput: bytes, timeLimit: float | None, memoryLimit: int | None
) -> CodeRunner.Result:
    try:
        result = CodeRunner().run(
            runCommand,
            inputContent=dataInput,
            timeLimit=timeLimit,
            memoryLimit=memoryLimit,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
    except OSError:
        return CodeRunner.Result(False, False, 0, b"", b"")
    return result



class CompilationError(Exception):
    def __init__(self, fileName: str, message: bytes, returnCode: int) -> None:
        self.fileName = fileName
        self.message = message
        self.returnCode = returnCode

    def getMessage(self) -> str:
        return self.message.decode()

    def __str__(self) -> str:
        return f"{self.fileName.capitalize()} compilation failed with return code {self.returnCode}."


class InputGenerationError(Exception):
    def __init__(self, clientID: str, returnCode: int) -> None:
        self.clientID = clientID
        self.returnCode = returnCode

    def __str__(self) -> str:
        return f"Input generation failed with return code {self.returnCode}."


class AnswerGenerationError(Exception):
    def __init__(self, clientID: str, returnCode: int) -> None:
        self.clientID = clientID
        self.returnCode = returnCode

    def __str__(self) -> str:
        return f"Answer generation failed with return code {self.returnCode}."



from autohack.core.util import *
import logging, pathlib, time, os


class Logger:
    def __init__(
        self,
        logFolder: pathlib.Path,
        logLevel: int = logging.WARNING,
        logTime: time.struct_time = time.localtime(),
    ) -> None:
        self.logFolder = logFolder
        self.logLevel = logLevel

        ensureDirExists(self.logFolder)

        self.logger = logging.getLogger("autohack")
        self.logger.setLevel(logLevel)

        self.logFilePath = self.logFolder / f"autohack-{formatTime(logTime)}.log"

        logFile = logging.FileHandler(self.logFilePath, encoding="utf-8")
        logFile.setLevel(logLevel)
        logFile.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s] - %(message)s")
        )

        self.logger.addHandler(logFile)

        self.logger.info(f'[logger] Log file: "{self.logFilePath}"')
        self.logger.info(f"[logger] Log level: {logging.getLevelName(logLevel)}")
        self.logger.info("[logger] Logger initialized.")

    def getLogger(self) -> logging.Logger:
        return self.logger

    def getLogFilePath(self) -> pathlib.Path:
        return self.logFilePath
